"""
Developer Name: Veronica Porubsky
Developer ORCID: 0000-0001-7216-3368
Developer GitHub Username: vporubsky
Developer Email: verosky@uw.edu

File Creation Date:

Description:
"""
# General mports
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from scipy import stats
import pandas as pd
import os

# Bokeh imports
from bokeh.io import show, save, output_notebook
from bokeh.models import Range1d, Circle, MultiLine
from bokeh.plotting import figure
from bokeh.plotting import from_networkx
from bokeh.palettes import Blues8
from bokeh.transform import linear_cmap



def interactive_network(ca_graph_obj, graph=None, attributes=['degree', 'HITS', 'hubs', 'CPR', 'communities'],
                        adjust_node_size=5, adjust_size_by='degree', adjust_color_by='communities',
                        palette=Blues8,
                        hover_attributes=['degree', 'HITS', 'hubs', 'CPR', 'communities'], title=None,
                        show_plot=True, show_in_notebook=False, save_plot=False, save_path=None):
    """
    Generates an interactived Bokeh.io plot of the graph network.

    palette: a color palette which can be passed as a tuple: palette = ('grey', 'red', 'blue')

    """
    # initialize graph information
    cg = ca_graph_obj
    if graph is None:
        G = cg.get_network_graph()
    else:
        G = graph
    label_keys = list(map(str, list(cg.labels)))

    #  Build attributes dictionary
    attribute_dict = {}
    for attribute in attributes:
        if attribute == 'degree':
            # Compute the degree of each node and add attribute
            attribute_dict['degree'] = dict(nx.degree(G))
        elif attribute == 'HITS':
            # Add HITS attribute
            hub_list, hit_vals = cg.get_hubs(graph=G)
            attribute_dict['HITS'] = hit_vals
        elif attribute == 'hubs':
            # Add hubs attribute
            attribute_dict['hubs'] = {i: 1 if i in list(set(hub_list) & set(label_keys)) else 0 for i in
                                      label_keys}
        elif attribute == 'CPR':
            # Add correlated pairs attribute
            corr_pair = cg.get_correlated_pair_ratio(graph=G)
            attribute_dict['CPR'] = {i: j for i, j in zip(label_keys, corr_pair)}
        elif attribute == 'communities':
            # Add communities
            c = list(nx.algorithms.community.greedy_modularity_communities(G))
            sorted(c)
            community_id = {}
            for i in range(len(c)):
                for j in list(c[i]):
                    community_id[j] = i
            attribute_dict['communities'] = community_id
        else:
            raise AttributeError('Invalid attribute key entered.')

    # Set node attributes
    for key, value in attribute_dict.items():
        nx.set_node_attributes(G, name=key, values=value)

    # Adjusted node size
    if adjust_node_size is not None:
        # Adjust node size
        adjusted_node_size = dict(
            [(node, value + adjust_node_size) for node, value in attribute_dict[adjust_size_by].items()])
        nx.set_node_attributes(G, name='adjusted_node_size', values=adjusted_node_size)
        size_by_this_attribute = 'adjusted_node_size'

    # Adjust node color
    color_by_this_attribute = adjust_color_by

    # Pick a color palette — Blues8, Reds8, Purples8, Oranges8, Viridis8
    color_palette = palette

    # Establish which categories will appear when hovering over each node
    HOVER_TOOLTIPS = [("Neuron", "@index")]
    for value in hover_attributes:
        HOVER_TOOLTIPS.append((value, "@" + value))

    # Create a plot with set dimensions, toolbar, and title
    plot = figure(tooltips=HOVER_TOOLTIPS,
                  tools="pan,wheel_zoom,save,reset", active_scroll='wheel_zoom',
                  x_range=Range1d(-10.1, 10.1), y_range=Range1d(-10.1, 10.1), title=title)

    # Create a network graph object
    position = nx.spring_layout(G)
    network_graph = from_networkx(G, position, scale=10, center=(0, 0))

    # Set node sizes and colors according to node degree (color as spectrum of color palette)
    minimum_value_color = min(network_graph.node_renderer.data_source.data[color_by_this_attribute])
    maximum_value_color = max(network_graph.node_renderer.data_source.data[color_by_this_attribute])
    network_graph.node_renderer.glyph = Circle(size=size_by_this_attribute,
                                               fill_color=linear_cmap(color_by_this_attribute, color_palette,
                                                                      minimum_value_color, maximum_value_color))

    # Set edge opacity and width
    network_graph.edge_renderer.glyph = MultiLine(line_alpha=0.5, line_width=1)

    plot.renderers.append(network_graph)
    if show_in_notebook:
        output_notebook()
        show(plot)
    elif show_plot:
        show(plot)

    if save_plot:
        if save_path is not None:
            save(plot, filename=save_path)
        else:
            save(plot, filename=os.path.join(os.getcwd(), f"bokeh_graph_visualization.html"))


def plot_CDF(data=None, color='black', marker='o', x_label='', y_label='CDF', show_plot=False):
    """
    Plots the cumulative distribution function of the provided list of data.

    :param data: list of float values
    :param color: str matplotlib color style
    :param marker: str matplotlib marker style
    :param x_label: str
    :param y_label: str
    :param show_plot: bool
    """

    # sort the dataset in ascending order
    sorted_data = np.sort(data)

    # get the cdf values of dataset
    cdf = np.arange(len(data)) / float(len(data))

    # plotting
    plt.plot(sorted_data, cdf, color=color, marker=marker)
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    if show_plot:
        plt.show()


def plot_CDF_compare_two_samples(data_list=None, color_list=['black', 'black'], marker='o', x_label='',
                                 y_label='CDF', show_plot=False):
    """
    Plots the cumulative distribution function of the provided datasets and prints the associated P-value for assessing
    the Kolmogorov-Smirnov distance between the distributions.

    :param data_list: list of lists containing float values to compare with KS-test
    :param color_list: list of str containing matplotlib color styles
    :param marker: str matplotlib marker style
    :param x_label: str
    :param y_label: str
    :param show_plot: bool
    """

    # Evaluate KS-test statistic
    stat_level = stats.ks_2samp(data_list[0], data_list[1])

    for idx, data in enumerate(data_list):
        # sort the dataset in ascending order
        sorted_data = np.sort(data)

        # get the cdf values of dataset
        cdf = np.arange(len(data)) / float(len(data))

        # plotting
        plt.plot(sorted_data, cdf, color=color_list[idx], marker=marker)

    plt.ylabel(y_label)
    plt.xlabel(x_label)
    plt.title(f'P value: {stat_level.pvalue:.2e}')

    if show_plot:
        plt.show()



def plot_histograms(data_list=None, color_list=['black', 'black'], x_label='',
                    y_label='count', bin_size=20, show_plot=True):
    """
    Plots histograms of the provided data and prints the associated P-value for assessing
    the Kolmogorov-Smirnov distance between the distributions.

    :param data_list: list of lists containing float values to compare with KS-test
    :param color_list: list of str containing matplotlib color styles
    :param marker: str matplotlib marker style
    :param x_label: str
    :param y_label: str
    :param show_plot: bool
    """

    # Evaluate KS-test statistic
    stat_level = stats.ks_2samp(data_list[0], data_list[1])

    for idx, data in enumerate(data_list):
        # plotting
        plt.hist(data, color=color_list[idx], bins=bin_size, alpha=0.4)

    plt.ylabel(y_label)
    plt.xlabel(x_label)
    plt.title(f'P value: {stat_level.pvalue:.2e}')

    if show_plot:
        plt.show()


def plot_matched_data(sample_1, sample_2, labels, value_label=None, colors=['grey', 'grey'], show_plot=True):
    """
    Plots two samples of matched data. Each sample will be plotted as points stacked vertically within condition.
    Lines will be drawn to connect the matching pairs.

    """
    # Put into dataframe
    df = pd.DataFrame({labels[0]: sample_1, labels[1]: sample_2})
    data = pd.melt(df)

    # Plot
    fig, ax = plt.subplots()
    sns.swarmplot(data=data, x='variable', y='value', ax=ax, size=0)

    # Find idx0 and idx1 by inspecting the elements returned from ax.get_children()
    idx0 = 0
    idx1 = 1
    locs1 = ax.get_children()[idx0].get_offsets()
    locs2 = ax.get_children()[idx1].get_offsets()

    y_all = np.zeros(2)
    for i in range(locs1.shape[0]):
        x = [locs1[i, 0], locs2[i, 0]]
        y = [locs1[i, 1], locs2[i, 1]]
        ax.plot(x, y, color='lightgrey', linewidth=0.5)
        ax.plot(locs1[i, 0], locs1[i, 1], '.', color=colors[0])
        ax.plot(locs2[i, 0], locs2[i, 1], '.', color=colors[1])
        data = [locs1[:, 1], locs2[:, 1]]
        ax.boxplot(data, positions=[0, 1], capprops=dict(linewidth=0.5, color='k'),
                   whiskerprops=dict(linewidth=0.5, color='k'),
                   boxprops=dict(linewidth=0.5, color='k'),
                   medianprops=dict(color='k'))
        plt.xticks([])
        y_all = np.vstack((y_all, y))
    if value_label is not None:
        plt.ylabel(value_label)
    plt.xlabel(f'P-value = {stats.ttest_rel(sample_1, sample_2).pvalue:.3}')

    if show_plot:
        plt.show()